#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf.h>
#include <boost/math/special_functions/binomial.hpp>

// birth process that manipulates cell-type abundances-vector
void birth(unsigned long long *Nvec, unsigned int i, unsigned int n, double u, gsl_rng * rng1) {
  // std::cout << "birth of type "<< i << '\n';  // #XXX testing
  Nvec[i]--; // parental cell is deleted
  // Segregate plasmids to daughter cells
  // Alternative 3:
  // Cell receive n1=n/2 and n2=n/2 plasmids for n even,
  //              n1=(n+1)/2 and n2=(n-1)/2 plasmids for n odd.

  int n1;
  int n2;
  if (n%2==0) { // n is even
    n1=n/2; // see above
    n2=n/2;
  } else {
    n1=(n+1)/2; // see above
    n2=(n-1)/2;
  }

  // mutated plasmids of daughter 1 after segregation before replication
  int i1=gsl_ran_hypergeometric(rng1,i,n-i,n1);
  // mutated plasmids of daughter 2 after segregation before replication
  int i2=i-i1;

  // std::cout << "final plasmid segregation "<< "\t\ti1="<<i1<<"\tn1="<<n1<< "\ti2="<<i2<<"\tn2="<<n2<<  '\n'; // XXX testing

  int k; // drawn plasmid-type

  // plasmid replication in daughter 1
  if (n1>0) { // check if daughter has plasmids
    int j1=i1; // declare number of mutated plasmids in daughter
    if (n>n1) { // check if daughter is not "full" (plasmid replication needed)
      if (i1==n1) { // check if daughter only bears mutated plasmids
        j1=n; // daughter after replication full of mutated plasmids
        n1=n;
      }  else {
        while (n1<n) {
          k=gsl_ran_binomial(rng1, (float)j1/n1,1); // draw one ball from urn
          j1+=k; // add mutant plasmid if mutant was drawn
          if (k==0) {
            k=gsl_ran_binomial(rng1, u, 1);
            j1+=k;
            // if (k==1) std::cout << "mutation" << '\n'; //XXX testing
          } // if wild-type: mutate
          n1++; // increase total plasmid number
        }
      }
    }
    // std::cout << "final plasmid replication in daughter 1    "<< "\tj1="<<j1<<"\tn1="<<n1<<  '\n'; //XXX testing
    Nvec[j1]++; // daughter cell added
  }

  // plasmid replication in daughter 2
  // plasmid replication in daughter 1
  if (n2>0) { // check if daughter has plasmids
    int j2=i2; // declare number of mutated plasmids in daughter
    if (n>n2) { // check if daughter is not "full" (plasmid replication needed)
      if (i2==n2) { // check if daughter only bears mutated plasmids
        j2=n; // daughter after replication full of mutated plasmids
        n2=n;
      } else {
        while (n2<n) {
          k=gsl_ran_binomial(rng1, (float)j2/n2,1); // draw one ball from urn
          j2+=k; // add mutant plasmid if mutant was drawn
          if (k==0) {
            k=gsl_ran_binomial(rng1, u, 1);
            j2+=k;
            // if (k==1) std::cout << "mutation" << '\n';
          } // if wild-type: mutate
          n2++; // increase total plasmid number
        }
      }
    }
    Nvec[j2]++; // daughter cell added
    // std::cout << "final plasmid replication in daughter 2    "<< "\tj2="<<j2<<"\tn2="<<n2<<  '\n'; //XXX testing
  }

}

// death process that manipulates cell-type abundances-vector
void death(unsigned long long *Nvec, unsigned int i, unsigned int n) {
  // std::cout << "death of type "<< i << '\n';  // #XXX testing
  Nvec[i]--;
}
