#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

// birth process that manipulates cell-type abundances-vector
void birth(unsigned long long *Nvec, unsigned int i, unsigned int n, double u, gsl_rng * rng1) {
  // Replicate plasmids, then segregate plasmids to daughter cells
  int itrans=i*2+gsl_ran_binomial(rng1,u,n-i); // number of mutated
    // after replication of each plasmid
  int j1=gsl_ran_hypergeometric(rng1,itrans,2*n-itrans,n); // mutated plasmids
    // in daughter 1
  int j2=itrans-j1; // mutated plasmids in dau. 2
  // update the array MP
  Nvec[i]--; // parental cell is deleted
  Nvec[j1]++; // daughter cells are added
  Nvec[j2]++;
  // printf("%d->%d+%d ", i,j1,j2 ); //XXX testing
}

// death process that manipulates cell-type abundances-vector
void death(unsigned long long *Nvec, unsigned int i, unsigned int n) {
  Nvec[i]--;
}

// replace process : one individual of type i divides. daughter 1 replaces
// parental cell and daughter 2 replaces one other cells


void birthandreplace(unsigned long long *Nvec, unsigned int i, unsigned int n, double u, gsl_rng * rng1) {

  // aux. variables for determining which cell is replaced
  double a[n+1];
  double sum_a_i; //
  double a0; //
  unsigned int r;
  double r2;   // random number variable

  // Replicate plasmids, then segregate plasmids to daughter cells
  int itrans=i*2+gsl_ran_binomial(rng1,u,n-i); // number of mutated
    // after replication of each plasmid
  int j1=gsl_ran_hypergeometric(rng1,itrans,2*n-itrans,n); // mutated plasmids
    // in daughter 1
  int j2=itrans-j1; // mutated plasmids in dau. 2
  // update the array MP
  Nvec[i]--; // parental cell is deleted
  // determine which cell is replaced
  for (int i = 0; i < (n+1); i++) a[i]=Nvec[i];
  a0=0.0; for (auto& ai : a) a0 += ai; // sum of a[] of all processes
  r2=gsl_rng_uniform(rng1);
  sum_a_i=0.0;r=0;
  do {sum_a_i += a[r];
    if (r2*a0<=sum_a_i) break;
    r++;}
    while(r<n+1);
  // printf("replacement of i=r=%d , daughter cells: %d, %d; ",r,j1,j2); // #XXX testing
  i=r;
  Nvec[i]--; // cell is deleted (replaced random cell)
  Nvec[j1]++; // daughter cell 1 is added (replacing parental cell)
  Nvec[j2]++; // daughter cell 2 is added (replacing random cell)
}
